#include <v_lib.h>

/* ==================================================================== */

#define	SCAN_POWER	8
#define	SCAN_WIDTH	((u32_t)1 << SCAN_POWER)

u32_t	image_width;	/* currently must be equal to SCAN_WIDTH */
u32_t	image_height;	/* currently must not be greater than SCAN_WIDTH */
vec_t	gray_pixels[SCAN_WIDTH][SCAN_WIDTH];
vec_t	*pix_rows[SCAN_WIDTH];
matrix_t *mask = NULL;

matrix_t pixmat =
{
	SCAN_WIDTH, SCAN_WIDTH,
	pix_rows,
	gray_pixels
};

void init_pixmat()
{
	int i;
	pixmat.cell[0] = pixmat.data;
	for(i=1;i<SCAN_WIDTH;i++)
		pixmat.cell[i] = pixmat.cell[i-1] + SCAN_WIDTH;
}
		
vec_t	freq_ccg[SCAN_WIDTH];
s32_t	quant = 128;

vboolean compensate = false;
vboolean quantize = false;
vec_t	pix_scale, pix_offset;

void calc_scale_and_offset(void)
{
	vec_t min = (vec_t)MAXFLOAT, max = -(vec_t)MAXFLOAT;
	u32_t count, line_num;
	for(line_num=0;line_num<image_height;line_num++)
	{
		vec_t *src = gray_pixels[line_num];
		for(count=0;count < SCAN_WIDTH;count++)
		{
			if(src[count] < min)
				min = src[count];
			if(src[count] > max)
				max = src[count];
		}
	}
	if(IsZero(max-min))
	{
		conlPuts(LOG_WARNING,
			"WARNING: IDFT result with zero contrast, luma compensation disabled\n");
		pix_offset = 128;
		pix_scale = 0;
		return;
	}
	pix_scale = (vec_t)256.0 / (max - min);
	pix_offset =  - min * pix_scale;
}

void scanlines_to_array(u8_t *pixels,u32_t channel)
{
	u32_t count, line_num;
	u32_t stride = SCAN_WIDTH * 4;
	u8_t *ptr, *endptr;
	vec_t *dst;
	for(line_num=0;line_num<image_height;line_num++)
	{
		ptr = pixels + stride * line_num;
		endptr = ptr + stride;
		dst = gray_pixels[line_num];
		for(count=0;ptr < endptr;ptr += 4,count++)
			dst[count] = ptr[channel];
	}
}

u8_t clamp(vec_t pixval)
{
	if(quantize)
	{
		if(pixval < quant)
			return 0;
		else
			return 255;
	}
	else
	{
		if(pixval < 0)
			return 0;
		if(pixval > 255)
			return 255;
		return (u8_t)(pixval + (vec_t)0.5);
	}
}

void array_to_scanlines(u8_t *pixels,u32_t channel)
{
	u32_t count, line_num;
	u32_t stride = SCAN_WIDTH * 4;
	u8_t *ptr, *endptr;
	vec_t *src;
	if(compensate)
		calc_scale_and_offset();
	else
	{
		pix_scale = 1;
		pix_offset = 0;
	}
	for(line_num=0;line_num<image_height;line_num++)
	{
		ptr = pixels + stride * line_num;
		endptr = ptr + stride;
		src = gray_pixels[line_num];
		for(count=0;ptr < endptr;ptr += 4,count++)
			ptr[channel] = clamp(pix_scale * src[count] + pix_offset);
	}
}

void fft_forward(u32_t line_num)
{
	fftRealNip(freq_ccg,pixmat.cell[line_num],SCAN_POWER,true);
}

void fft_inverse(u32_t line_num)
{
	fftRealNip(pixmat.cell[line_num],freq_ccg,SCAN_POWER,false);
}

void fft2d(matrix_t *m,u32_t size_power,vboolean forward)
{
	u32_t i;
	matrix_t *m2;
	V_ASSERT((m->cols == m->rows) && (m->rows == (u32_t)(1 << size_power)));
	m2 = matDuplicate(m);
	for(i=0;i<m->rows;i++)
		fftRealNip(m->cell[i],m2->cell[i],size_power,forward);
	matTranspose(m2,m);
	for(i=0;i<m->rows;i++)
		fftRealNip(m->cell[i],m2->cell[i],size_power,forward);
	memFree(m2);
}

void zero_freq(vec_t *afreq,s32_t lo_pass,s32_t hi_pass)
{
	u32_t hlen;
	/*
	 * Discard frequencies below 'lo_pass'
	 */
	if(lo_pass)
		memZero(afreq,lo_pass * 2 * sizeof(vec_t));
	/*
	 * Discard frequencies above 'hi_pass'
	 */
	hi_pass <<= 1;
	hlen = SCAN_WIDTH - hi_pass;
	if(hlen)
		memZero(afreq + hi_pass,hlen * sizeof(vec_t));
}

void zero2d(matrix_t *m,s32_t lo_pass,s32_t hi_pass)
{
	u32_t i;
	matrix_t *m2 = matAlloc(m->rows,m->cols);
	for(i=0;i<image_height;i++)
		zero_freq(m->cell[i],lo_pass,hi_pass);
	matTranspose(m2,m);
	for(i=0;i<image_height;i++)
		zero_freq(m2->cell[i],lo_pass,hi_pass);
	matTranspose(m,m2);
	memFree(m2);
}

void band_pass(u8_p pixels,s32_t lo_pass,s32_t hi_pass)
{
	u32_t channel;
	for(channel=0;channel<3;channel++)
	{
		scanlines_to_array(pixels,channel);
		fft2d(&pixmat,SCAN_POWER,true);
		zero2d(&pixmat,lo_pass,hi_pass);
		fft2d(&pixmat,SCAN_POWER,false);
		array_to_scanlines(pixels,channel);
	}
}

void masked_zero(matrix_t *m,matrix_t *mask)
{
	u32_t i,j;
	for(i=0;i<m->rows;i++)
		for(j=0;j<m->cols;j++)
			m->cell[i][j] *= mask->cell[i][j];
}

void mask_pass(u8_p pixels,matrix_t *mask)
{
	u32_t channel;
	for(channel=0;channel<3;channel++)
	{
		scanlines_to_array(pixels,channel);
		fft2d(&pixmat,SCAN_POWER,true);
		masked_zero(&pixmat,mask);
		fft2d(&pixmat,SCAN_POWER,false);
		array_to_scanlines(pixels,channel);
	}
}

/* ==================================================================== */

void usage(char *name)
{
	puts("Band-pass image filtering by FDFT/IDFT. Version 0.1 by Vladimir A. Smirnov.\n");
	printf("Usage: %s <infile> (<low bound> <high bound> | mask <maskfile>) [outfile] [outmethod], where:\n",name);
	puts("\t<infile> - name of image file to process (BMP, TGA or JPG);");
	puts("\t<low freq bound>, <high freq bound> - bounds in wave-number domain;");
	puts("\t[outfile] - optional name of output file (default is 'out.bmp').");
	printf("Width of the input bitmap must be %u, height must not be greater than %u.\n",SCAN_WIDTH,SCAN_WIDTH);
	printf("Bounds in wave-number space must be in range [0..%u].\n",SCAN_WIDTH>>1);
	puts("Last parameter must be either 'clamp' (clamp results of IDFT in range 0..255),");
	puts("or 'compensate' (increase contrast such that luma occupy range from 0 to 255),");
	puts("or 'quant' (same as compensate, but produce mono output); default is 'clamp'.");
	puts("If 'outmethod' is 'quant', it's possible to specify threshold (ex.: quant=77).");
	puts("KNOWN BUGS: Probably, a lot.");
}

u8_p LoadImage(char *name,u32_p w_ret,u32_p h_ret)
{
	char ext[_MAX_PATH];
	fsSplitPath(name,NULL,NULL,ext);
	if(!strcasecmp(ext,"bmp"))
		return bmpLoad(name,w_ret,h_ret);
	if(!strcasecmp(ext,"tga"))
		return tgaLoad(name,w_ret,h_ret);
	if(!strcasecmp(ext,"jpg"))
		return jpgLoad(name,w_ret,h_ret);
	conlPrintf(LOG_ERROR,"LoadImage(%s): unrecognized format.\n",name);
	return NULL;
}

matrix_p LoadMask(char *name)
{
	u8_p ptr, pixbuf;
	matrix_p mask;
	u32_t w, h, i, j;
	pixbuf = LoadImage(name,&w, &h);
	if(!pixbuf)
		return NULL;
	if(w != SCAN_WIDTH || h != SCAN_WIDTH)
	{
		memFree(pixbuf);
		return NULL;
	}
	mask = matAlloc(w,h);
	ptr = pixbuf;	
	for(i=0;i<mask->rows;i++)
	{
		for(j=0;j<mask->cols;j++,ptr+=4)
			mask->cell[i][j] = (vec_t)((u32_t)ptr[0]+(u32_t)ptr[1]+(u32_t)ptr[2]) / (vec_t)(0xFF * 3);
	}
	return mask;
}

void SaveImage(char *name,u32_t w,u32_t h,u8_p pixels)
{
	char ext[_MAX_PATH];
	fsSplitPath(name,NULL,NULL,ext);
	if(!strcasecmp(ext,"bmp"))
		bmpSave(name,w,h,pixels);
	else if(!strcasecmp(ext,"tga"))
		tgaSave(name,w,h,pixels);
	else
	{
		conlPrintf(LOG_WARNING,"SaveImage(%s): unrecognized format, assuming TARGA.\n",name);
		tgaSave(name,w,h,pixels);
	}
}

vboolean set_clamp(char *arg)
{
	if(!strcmp(arg,"compensate"))
		compensate = true;
	else if(!strcmp(arg,"clamp"))
		compensate = false;
	else
	{
		char buf[64];
		memCopy(buf,arg,5);
		buf[5]=0;
		if(!strcmp(buf,"quant"))
		{
			compensate = true;
			quantize = true;
			if(arg[5] == '=')
				strToInt(arg+6,&quant);
		}
		else
			return false;
	}
	return true;
}

vboolean init(int argc,char *argv[],s32_t *l_ret,s32_t *h_ret,char **outname_ret)
{
	s32_t fmax = SCAN_WIDTH >> 1;
	init_pixmat();
	if(argc < 4 || argc > 6)
	{
		usage(argv[0]);
		return false;
	}
	if(argc == 4) {
		*outname_ret = NULL;
		compensate = false;
	} else if(argc == 5) {
		if(!set_clamp(argv[4]))
			*outname_ret = argv[4];
		else
			*outname_ret = NULL;
	}
	else /* argc == 6 */
	{
		u32_t no = 4;
		if(set_clamp(argv[4]))
			no = 5;
		else
		{
			if(!set_clamp(argv[5]))
			{
				conlPrintf(LOG_ERROR,
					"init(): either 'compensate' or 'clamp' got '%s'.\n"
					"For help, run %s without any arguments.\n",
					argv[5],argv[0]);
				return false;
			}
		}
		*outname_ret = argv[no];
	}
	if(strcasecmp(argv[2],"mask"))
	{
    	if(!strToInt(argv[2],l_ret) || !strToInt(argv[3],h_ret))
    	{
    		conlPrintf(LOG_ERROR,
    			"init(): expected two integers, got '%s' and '%s'.\n"
    			"For help, run %s without any arguments.\n",
    			argv[2],argv[3],argv[0]);
    		return false;
    	}
    	if(*l_ret < 0 || *l_ret > fmax || *h_ret < 0 || *h_ret > fmax)
    	{
    		conlPrintf(LOG_ERROR,
    			"init(): wrong frequency bounds '%d' and '%d'.\n"
    			"For help, run %s without any arguments.\n",
    			*l_ret,*h_ret,argv[0]);
    		return false;
    	}
    }
   	else
    {
    	mask = LoadMask(argv[3]);
    	if(mask == NULL)
    		return false;
    }
	return true;
}

int main(int argc,char *argv[])
{
	s32_t lo_pass,hi_pass;
	u8_t *pixels;
	char *outname;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	if(!init(argc,argv,&lo_pass,&hi_pass,&outname))
		return -1;
	if(!outname)
		outname = "out.bmp";
	pixels = LoadImage(argv[1],&image_width,&image_height);
	if(!pixels)
	{
		if(mask)
			memFree(mask);
		return -1;
	}
	if(image_width != SCAN_WIDTH)
	{
		conlPrintf(LOG_ERROR,"Bitmap width must be equal to %u, got %u\n",SCAN_WIDTH,image_width);
		memFree(pixels);
		if(mask)
			memFree(mask);
		return -1;
	}
	if(image_height > SCAN_WIDTH)
	{
		conlPrintf(LOG_ERROR,"Bitmap height must be less than %u, got %u\n",SCAN_WIDTH,image_height);
		memFree(pixels);
		if(mask)
			memFree(mask);
		return -1;
	}
	if(mask)
		mask_pass(pixels,mask);
	else
		band_pass(pixels,lo_pass,hi_pass);
	SaveImage(outname,image_width,image_height,pixels);
	memFree(pixels);
	if(mask)
		memFree(mask);
	return 0;
}
